import argparse
import sys

from utils.benchmark.benchmark_runner import BenchmarkRunner
from utils.definitions.list_functions import AllFunctionDefinitionPrinter
from cythonpowered import VERSION


TITLE = f"""
               _   _                                                      _ 
     ___ _   _| |_| |__   ___  _ __  _ __   _____      _____ _ __ ___  __| |
    / __| | | | __| '_ \ / _ \| '_ \| '_ \ / _ \ \ /\ / / _ \ '__/ _ \/ _` |
   | (__| |_| | |_| | | | (_) | | | | |_) | (_) \ V  V /  __/ | |  __/ (_| |
    \___|\__, |\__|_| |_|\___/|_| |_| .__/ \___/ \_/\_/ \___|_|  \___|\__,_|
         |___/                      |_|                                     
                                                                  ver. {VERSION}
"""


def run_benchmark():
    print(TITLE, flush=True)
    BenchmarkRunner()


def list_functions():
    print(TITLE, flush=True)
    AllFunctionDefinitionPrinter()


def print_version():
    print(VERSION)


def main():
    parser = argparse.ArgumentParser(
        prog="cythonpowered",
        description="Utilities for the cythonpowered Python library",
    )

    parser.add_argument("-b", "--benchmark", help="run benchmark", action="store_true")
    parser.add_argument("-l", "--list", help="list functions", action="store_true")
    parser.add_argument("-v", "--version", help="print version", action="store_true")
    args = parser.parse_args()

    passed_args = [
        arg for arg in args.__dict__ if args.__dict__[arg] not in [False, None]
    ]
    if len(passed_args) == 0:
        parser.print_help()

    if args.benchmark is True:
        run_benchmark()
        sys.exit(0)

    if args.list is True:
        list_functions()
        sys.exit(0)

    if args.version is True:
        print_version()
        sys.exit(0)


if __name__ == "__main__":
    main()



# -----------------------------------------------------------------------------
# Import functions from stdlib.h C library
cdef extern from "stdlib.h":
    void srand(unsigned int sd)
    unsigned int rand()
    unsigned long clock()
    double drand48()
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
cpdef double random():
# Returns a random float (double) between 0 and 1
# Replacement for random.random()
    return drand48()


cpdef list n_random(unsigned int n):
    # Returns a n-element list of random floats (double) between 0 and 1
    cdef unsigned int i
    return [drand48() for i in range(n)]
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
cpdef int randint(int a, int b):
# Returns a random integer in range [a, b]
# Replacement for random.randint()
    return rand() % (b - a + 1) + a


cpdef list n_randint(int a, int b, unsigned int n):
    # Returns a n-element list of random integers in range [a, b]
    cdef unsigned int i
    return [randint(a, b) for i in range(n)]
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
cpdef double uniform(double a, double b):
    # Returns a random float (double) between in range [a, b]
    # Replacement for random.uniform()
    return a + (b - a) * drand48()
    

cpdef list n_uniform(double a, double b, unsigned int n):
    # Returns a n-element list of random floats (double) in range [a, b]
    cdef unsigned int i
    return [uniform(a, b) for i in range(n)]
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
cpdef choice(list population):
    # Returns a random element from a given list
    # Replacement for random.choice()
    cdef unsigned int size = len(population)
    return population[rand() % size]


cpdef list choices(list population, unsigned int k=1):
    # Given a list, returns a selection (list) of n random elements
    # Replacement for random.choices()
    cdef unsigned int i
    cdef unsigned int size = len(population)
    srand(clock())
    return [population[rand() % size] for i in range(k)]
# -----------------------------------------------------------------------------



from utils.definitions._base import BaseFunctionDefinition, REPLACEMENT
import random as py_random
import cythonpowered.random as cy_random


# =============================================================================
class PythonRandomRandomDef(BaseFunctionDefinition):
    function = py_random.random
    reference = "random.random"


class CythonRandomRandomDef(BaseFunctionDefinition):
    function = cy_random.random
    reference = "cythonpowered.random.random"
    usage = REPLACEMENT


class CythonRandomNRandomDef(BaseFunctionDefinition):
    function = cy_random.n_random
    reference = "cythonpowered.random.n_random"
    usage = "n_random(k) is equivalent to [random() for i in range(k)]"


# =============================================================================
class PythonRandomRandintDef(BaseFunctionDefinition):
    function = py_random.randint
    reference = "random.randint"


class CythonRandomRandintDef(BaseFunctionDefinition):
    function = cy_random.randint
    reference = "cythonpowered.random.randint"
    usage = REPLACEMENT


class CythonRandomNRandintDef(BaseFunctionDefinition):
    function = cy_random.n_randint
    reference = "cythonpowered.random.n_randint"
    usage = "n_randint(a, b, k) is equivalent to [randint(a, b) for i in range(k)]"


# =============================================================================
class PythonRandomUniformDef(BaseFunctionDefinition):
    function = py_random.uniform
    reference = "random.uniform"


class CythonRandomUniformDef(BaseFunctionDefinition):
    function = cy_random.uniform
    reference = "cythonpowered.random.uniform"
    usage = REPLACEMENT


class CythonRandomNUniformDef(BaseFunctionDefinition):
    function = cy_random.n_uniform
    reference = "cythonpowered.random.n_uniform"
    usage = "n_uniform(a, b, k) is equivalent to [uniform(a, b) for i in range(k)]"


# =============================================================================
class PythonRandomChoiceDef(BaseFunctionDefinition):
    function = py_random.choice
    reference = "random.choice"


class CythonRandomChoiceDef(BaseFunctionDefinition):
    function = cy_random.choice
    reference = "cythonpowered.random.choice"
    usage = REPLACEMENT


# =============================================================================
class PythonRandomChoicesDef(BaseFunctionDefinition):
    function = py_random.choices
    reference = "random.choices"


class CythonRandomChoicesDef(BaseFunctionDefinition):
    function = cy_random.choices
    reference = "cythonpowered.random.choices"
    usage = "Drop-in replacement, only supports the 'k' keyword argument"


# =============================================================================


RANDOM_DEFINITION_PAIRS = [
    [PythonRandomRandomDef, CythonRandomRandomDef],
    [PythonRandomRandomDef, CythonRandomNRandomDef],
    [PythonRandomRandintDef, CythonRandomRandintDef],
    [PythonRandomRandintDef, CythonRandomNRandintDef],
    [PythonRandomUniformDef, CythonRandomUniformDef],
    [PythonRandomUniformDef, CythonRandomNUniformDef],
    [PythonRandomChoiceDef, CythonRandomChoiceDef],
    [PythonRandomChoicesDef, CythonRandomChoicesDef],
]



from cpuinfo import get_cpu_info
import platform
import psutil
from typing import Optional

from utils.benchmark._random import RandomBenchmark


class BenchmarkRunner:

    MODULE_BENCHMARKS = [RandomBenchmark]

    def __init__(self) -> None:
        sys_info = self.get_system_info()
        sys_info_pretty = self.prettify_dict(sys_info)
        for line in sys_info_pretty:
            self.log(line)
        self.run_module_benchmarks()

    def log(self, msg, end: Optional[str] = None) -> None:
        print(msg, end=end, flush=True)

    def get_system_info(self) -> dict:

        # CPU info
        default_unsupported_param = "Unsupported parameter for this system"
        cpu = get_cpu_info()
        cpu_name = cpu.get("brand_raw", f"[{default_unsupported_param}]")
        cpu_freq = cpu.get("hz_advertised_friendly", f"[{default_unsupported_param}]")
        cpu_cores = psutil.cpu_count(logical=False)
        cpu_threads = psutil.cpu_count(logical=True)
        arch = platform.machine()

        # Memory info
        svmem = psutil.virtual_memory()
        ram_size_gb = svmem.total / (1024**3)
        ram = f"{ram_size_gb:.2f} GB"

        # OS info
        os_name = platform.system()
        os_release = platform.release()
        os = f"{os_name} {os_release}"

        # Python version
        py_version = platform.python_version()

        # Compiler info
        compiler = platform.python_compiler()

        sysinfo = {
            "CPU model": cpu_name,
            "CPU base frequency": cpu_freq,
            "CPU cores": cpu_cores,
            "CPU threads": cpu_threads,
            "Architecture": arch,
            "Memory (RAM)": ram,
            "Operating System": os,
            "Python version": py_version,
            "C compiler": compiler,
        }

        return sysinfo

    def prettify_dict(self, input_dict: dict) -> list:
        spaces = 4
        keys = [k for k in input_dict.keys()]
        max_len = max([len(k) for k in keys])
        prettified = [f"{k}:{' '*(max_len-len(k)+spaces)}{input_dict[k]}" for k in keys]
        return prettified

    def run_module_benchmarks(self):
        for module_benchmark in self.MODULE_BENCHMARKS:
            self.log("")
            module_benchmark()



from utils.benchmark._base import (
    BaseFunctionBenchmark,
    BaseModuleBenchmark,
)

from utils.definitions._random import (
    PythonRandomRandomDef,
    CythonRandomRandomDef,
    CythonRandomNRandomDef,
    PythonRandomRandintDef,
    CythonRandomRandintDef,
    CythonRandomNRandintDef,
    PythonRandomUniformDef,
    CythonRandomUniformDef,
    CythonRandomNUniformDef,
    PythonRandomChoiceDef,
    CythonRandomChoiceDef,
    PythonRandomChoicesDef,
    CythonRandomChoicesDef,
)

import cythonpowered.random as cy_random


class RandomBenchmarkDefinition(BaseFunctionBenchmark):
    python_function = PythonRandomRandomDef
    cython_function = CythonRandomRandomDef
    cython_n_function = CythonRandomNRandomDef


class RandintBenchmarkDefinition(BaseFunctionBenchmark):
    python_function = PythonRandomRandintDef
    cython_function = CythonRandomRandintDef
    cython_n_function = CythonRandomNRandintDef
    args = [-1000000, 1000000]


class UniformBenchmarkDefinition(BaseFunctionBenchmark):
    python_function = PythonRandomUniformDef
    cython_function = CythonRandomUniformDef
    cython_n_function = CythonRandomNUniformDef
    args = [-123456.789, 123456.789]


class ChoiceBenchmarkDefinition(BaseFunctionBenchmark):
    python_function = PythonRandomChoiceDef
    cython_function = CythonRandomChoiceDef
    args = [cy_random.n_randint(-100000, 100000, 10000)]


class ChoicesBenchmarkDefinition(BaseFunctionBenchmark):
    python_function = PythonRandomChoicesDef
    cython_function = CythonRandomChoicesDef
    args = [cy_random.n_randint(-100000, 100000, 10000)]
    kwargs = {"k": 100}
    runs = [1000, 10000, 100000]


class RandomBenchmark(BaseModuleBenchmark):
    MODULE = "random"
    BENCHMARKS = [
        RandomBenchmarkDefinition,
        RandintBenchmarkDefinition,
        UniformBenchmarkDefinition,
        ChoiceBenchmarkDefinition,
        ChoicesBenchmarkDefinition,
    ]



from utils.definitions._base import BaseFunctionDefinitionPrinter
from utils.definitions._random import RANDOM_DEFINITION_PAIRS


ALL_PAIRS = RANDOM_DEFINITION_PAIRS


class AllFunctionDefinitionPrinter(BaseFunctionDefinitionPrinter):
    FUNCTION_PAIRS = ALL_PAIRS



from typing import Optional, List
from time import time
from prettytable import PrettyTable

from cythonpowered import MODULES
from utils import SEPARATOR
from utils.definitions._base import BaseFunctionDefinition


DEFAULT_RUNS = [10000, 100000, 1000000]


class BaseFunctionBenchmark:
    python_function: BaseFunctionDefinition
    cython_function: BaseFunctionDefinition
    cython_n_function: Optional[BaseFunctionDefinition] = None
    args: list = []
    kwargs: dict = {}
    runs: Optional[List[int]] = DEFAULT_RUNS

    def __init__(self) -> None:
        self.benchmark_data = {}
        self.run()

    def log(self, msg, end: Optional[str] = None) -> None:
        print(msg, end=end, flush=True)

    def log_progress(self, progress: int, total: int):
        cython_n_func_ref = None
        if self.cython_n_function is not None:
            cython_n_func_ref = self.cython_n_function.reference

        if progress == 0:
            end = ""
        else:
            self.log("\r", end="")
            end = "\r"

        cython_n_func_msg = (
            " and [" + cython_n_func_ref + "]" if cython_n_func_ref is not None else ""
        )
        self.log(
            f"Comparing [{self.python_function.reference}] with [{self.cython_function.reference}]{cython_n_func_msg}... {(progress/total*100):.2f}%",
            end=end,
        )

    def run(self):
        python_func = self.python_function.function
        cython_func = self.cython_function.function
        cython_n_func = None

        if self.cython_n_function is not None:
            cython_n_func = self.cython_n_function.function

        args = self.args
        kwargs = self.kwargs
        runs = self.runs

        progress = 0
        total = 6 if cython_n_func is None else 9

        self.log_progress(progress, total)

        for run in runs:
            # Run Python function
            st = time()
            py_results = [python_func(*args, **kwargs) for i in range(run)]
            et = time()
            python_time = et - st

            progress += 1
            self.log_progress(progress, total)

            # Run cythonpowered function
            st = time()
            cy_results = [cython_func(*args, **kwargs) for i in range(run)]
            et = time()
            cython_time = et - st

            progress += 1
            self.log_progress(progress, total)

            # Run cythonpowered n_function
            cython_n_time = None
            if cython_n_func is not None:
                st = time()
                n_results = cython_n_func(*args, run, **kwargs)
                et = time()
                cython_n_time = et - st

                progress += 1
                self.log_progress(progress, total)

            self.benchmark_data[str(run)] = [python_time, cython_time, cython_n_time]

        self.log("")


class BaseModuleBenchmark:

    MODULE: Optional[str] = None
    BENCHMARKS: List[BaseFunctionBenchmark] = []

    def __init__(self) -> None:
        if self.MODULE not in MODULES:
            raise NotImplementedError(f"Unknown module [{self.MODULE}]")
        self.intro()
        self.results = [b() for b in self.BENCHMARKS]
        self.print_results()

    def log(self, msg, end: Optional[str] = None) -> None:
        print(msg, end=end, flush=True)

    def intro(self) -> None:
        module = self.MODULE
        defs = len(self.BENCHMARKS)
        self.log(SEPARATOR)
        self.log(
            f"Running benchmark for the [cythonpowered.{module}] module ({defs} benchmarks)..."
        )
        self.log(SEPARATOR)

    def format_number_magnitude(self, number) -> str:
        magnitudes = ["", "K", "M", "B"]
        i = 0
        while number // (1000**i) >= 1000 and i <= len(magnitudes):
            i += 1
        formatted = f"{int(number / (1000 ** i))}{magnitudes[i]}"
        return formatted

    def format_execution_times(self, times: List[float]) -> str:
        formatted = str(
            [
                f"{('{0:.' + str(4 - i) + 'f}').format(times[i])}"
                for i in range(len(times))
            ]
        ).replace("'", "")
        return formatted

    def format_factor(self, factor) -> str:
        _format = "{0:." + ("2" if factor < 10 else "1") + "f}"
        return _format.format(factor)

    def format_factors(
        self,
        times_1: List[float],
        times_2: List[float],
        inverted: bool = False,
    ) -> str:

        if inverted:
            formatted = str(
                [
                    self.format_factor(times_2[i] / times_1[i])
                    for i in range(len(times_2))
                ]
            ).replace("'", "")
        else:
            formatted = str(
                [
                    self.format_factor(times_1[i] / times_2[i])
                    for i in range(len(times_2))
                ]
            ).replace("'", "")

        return formatted

    def calculate_avg_factor(
        self,
        times_1: List[float],
        times_2: List[float],
        inverted: bool = False,
    ) -> str:
        if inverted is True:
            return sum(([times_1[i] / times_2[i] for i in range(len(times_1))])) / len(
                times_1
            )
        else:
            return sum([times_2[i] / times_1[i] for i in range(len(times_1))]) / len(
                times_2
            )

    def print_results(self) -> None:
        table = PrettyTable()
        table.field_names = [
            "Function name",
            "No. of runs",
            "Execution time (s)",
            "Time factor",
            "Speed factor",
            "Avg. speed factor",
        ]

        for r in self.results:

            items = [i[1] for i in r.benchmark_data.items()]

            cython_n_function = r.cython_n_function
            python_times = [t[0] for t in items]
            cython_times = [t[1] for t in items]
            if cython_n_function is not None:
                cython_n_times = [t[2] for t in items]

            runs = str([self.format_number_magnitude(n) for n in r.runs]).replace(
                "'", ""
            )

            table.add_row(
                [
                    f"[Python] {r.python_function.reference}",
                    runs,
                    self.format_execution_times(python_times),
                    "1.00",
                    "1.00",
                    "1.00",
                ]
            )
            table.add_row(
                [
                    r.cython_function.reference,
                    runs,
                    self.format_execution_times(cython_times),
                    self.format_factors(cython_times, python_times),
                    self.format_factors(cython_times, python_times, inverted=True),
                    self.format_factor(
                        self.calculate_avg_factor(cython_times, python_times)
                    ),
                ],
                divider=(cython_n_function is None),
            )

            if cython_n_function is not None:
                table.add_row(
                    [
                        r.cython_n_function.reference,
                        runs,
                        self.format_execution_times(cython_n_times),
                        self.format_factors(cython_n_times, python_times),
                        self.format_factors(
                            cython_n_times, python_times, inverted=True
                        ),
                        self.format_factor(
                            self.calculate_avg_factor(
                                cython_n_times, python_times, r.runs
                            )
                        ),
                    ],
                    divider=True,
                )

        self.log(table)



from typing import List, Optional, Type
from prettytable import PrettyTable

REPLACEMENT = "Drop-in replacement"


class BaseFunctionDefinition:
    function: callable
    reference: str
    usage: str = ""


class BaseFunctionDefinitionPrinter:

    FUNCTION_PAIRS: List[List[BaseFunctionDefinition]] = []

    def __init__(self) -> None:
        if len(self.FUNCTION_PAIRS) == 0:
            raise NotImplementedError("No pairs of function definitions provided")
        for pair in self.FUNCTION_PAIRS:
            if type(pair) != list:
                raise TypeError
            if len(pair) != 2:
                raise ValueError
        self.print_function_definitions()

    def log(self, msg, end: Optional[str] = None) -> None:
        print(msg, end=end, flush=True)

    def print_function_definitions(self) -> None:
        table = PrettyTable()
        table.field_names = [
            "#",
            "[cythonpowered] function",
            "Replaces [Python] function",
            "Usage / details",
        ]
        for f in table.field_names:
            table.align[f] = "r" if f == table.field_names[0] else "l"

        for i in range(len(self.FUNCTION_PAIRS)):
            table.add_row(
                [
                    i + 1,
                    self.FUNCTION_PAIRS[i][1].reference,
                    self.FUNCTION_PAIRS[i][0].reference,
                    self.FUNCTION_PAIRS[i][1].usage,
                ]
            )
        self.log(table)
